// Here begins Pyrex's Standard Shadowcasting Implementation
// (I copy this to lots of projects!)
export var shadowcast = function (
  [ox, oy]: [number, number],
  isBlocking: (xy: [number, number]) => boolean,
  markVisible: (xy: [number, number]) => void,
) {
  for (var i = 0; i < 4; i++) {
    var quadrant = new Quadrant(i, [ox, oy]);
    var reveal = function (xy: [number, number]) {
      markVisible(quadrant.transform(xy));
    };
    var isWall = function (xy: [number, number] | undefined) {
      if (xy == undefined) {
        return false;
      }
      return isBlocking(quadrant.transform(xy));
    };
    var isFloor = function (xy: [number, number] | undefined) {
      if (xy == undefined) {
        return false;
      }
      return !isBlocking(quadrant.transform(xy));
    };
    var scan = function (row: Row) {
      var prevXy: [number, number] | undefined;
      row.forEachTile((xy) => {
        if (isWall(xy) || isSymmetric(row, xy)) {
          reveal(xy);
        }
        if (isWall(prevXy) && isFloor(xy)) {
          row.startSlope = slope(xy);
        }
        if (isFloor(prevXy) && isWall(xy)) {
          var nextRow = row.next();
          nextRow.endSlope = slope(xy);
          scan(nextRow);
        }
        prevXy = xy;
      });
      if (isFloor(prevXy)) {
        scan(row.next());
      }
    };

    var firstRow = new Row(1, new Fraction(-1, 1), new Fraction(1, 1));
    scan(firstRow);
  }
};

class Quadrant {
  cardinal: number;
  ox: number;
  oy: number;

  constructor(cardinal: number, [ox, oy]: [number, number]) {
    this.cardinal = cardinal;
    this.ox = ox;
    this.oy = oy;
  }

  transform([row, col]: [number, number]): [number, number] {
    switch (this.cardinal) {
      case 0:
        return [this.ox + col, this.oy - row];
      case 2:
        return [this.ox + col, this.oy + row];
      case 1:
        return [this.ox + row, this.oy + col];
      case 3:
        return [this.ox - row, this.oy + col];
      default:
        throw new Error("invalid cardinal");
    }
  }
}

class Row {
  depth: number;
  startSlope: Fraction;
  endSlope: Fraction;

  constructor(depth: number, startSlope: Fraction, endSlope: Fraction) {
    this.depth = depth;
    this.startSlope = startSlope;
    this.endSlope = endSlope;
  }

  forEachTile(cb: (xy: [number, number]) => void) {
    var minCol = roundTiesUp(this.startSlope.scale(this.depth));
    var maxCol = roundTiesDown(this.endSlope.scale(this.depth));
    for (var col = minCol; col <= maxCol; col++) {
      cb([this.depth, col]);
    }
  }
  next(): Row {
    return new Row(this.depth + 1, this.startSlope, this.endSlope);
  }
}

class Fraction {
  numerator: number;
  denominator: number;

  constructor(numerator: number, denominator: number) {
    this.numerator = numerator;
    this.denominator = denominator;
  }

  scale(n: number): Fraction {
    return new Fraction(this.numerator * n, this.denominator);
  }

  toDouble(): number {
    return this.numerator / this.denominator;
  }
}

var slope = function ([rowDepth, col]: [number, number]): Fraction {
  return new Fraction(2 * col - 1, 2 * rowDepth);
};

var isSymmetric = function (row: Row, [_, col]: [number, number]) {
  return (
    col >= row.startSlope.scale(row.depth).toDouble() &&
    col <= row.endSlope.scale(row.depth).toDouble()
  );
};

var roundTiesUp = function (n: Fraction) {
  return Math.floor(n.toDouble() + 0.5);
};

var roundTiesDown = function (n: Fraction) {
  return Math.ceil(n.toDouble() - 0.5);
};
